@@ -1,81 +1,29 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import sys
 from time import sleep
 from threading import Event
 from collections import deque
+from queue import Queue
 
 from xpra.make_thread import start_thread
-from xpra.os_util import Queue, monotonic_time
-from xpra.util import merge_dicts, flatten_dict, notypedict, envbool, envint, typedict, AtomicInteger
+from xpra.os_util import monotonic_time
+from xpra.util import notypedict, envbool, envint, typedict, AtomicInteger
 from xpra.net.compression import compressed_wrapper, Compressed
 from xpra.server.source.source_stats import GlobalPerformanceStatistics
-from xpra.server.source.clientinfo_mixin import ClientInfoMixin
-from xpra.server import server_features
+from xpra.server.source.stub_source_mixin import StubSourceMixin
 from xpra.log import Logger
 
-CC_BASES = [ClientInfoMixin]
-#TODO: notifications mixin
-if server_features.clipboard:
-    from xpra.server.source.clipboard_connection import ClipboardConnection
-    CC_BASES.append(ClipboardConnection)
-if server_features.audio:
-    from xpra.server.source.audio_mixin import AudioMixin
-    CC_BASES.append(AudioMixin)
-if server_features.webcam:
-    from xpra.server.source.webcam_mixin import WebcamMixin
-    CC_BASES.append(WebcamMixin)
-if server_features.fileprint:
-    from xpra.server.source.fileprint_mixin import FilePrintMixin
-    CC_BASES.append(FilePrintMixin)
-if server_features.mmap:
-    from xpra.server.source.mmap_connection import MMAP_Connection
-    CC_BASES.append(MMAP_Connection)
-if server_features.input_devices:
-    from xpra.server.source.input_mixin import InputMixin
-    CC_BASES.append(InputMixin)
-if server_features.dbus:
-    from xpra.server.source.dbus_mixin import DBUS_Mixin
-    CC_BASES.append(DBUS_Mixin)
-if server_features.network_state:
-    from xpra.server.source.networkstate_mixin import NetworkStateMixin
-    CC_BASES.append(NetworkStateMixin)
-if server_features.display:
-    from xpra.server.source.clientdisplay_mixin import ClientDisplayMixin
-    CC_BASES.append(ClientDisplayMixin)
-if server_features.windows:
-    from xpra.server.source.windows_mixin import WindowsMixin
-    CC_BASES.append(WindowsMixin)
-    #must be after windows mixin so it can assume "self.send_windows" is set
-    if server_features.encoding:
-        from xpra.server.source.encodings_mixin import EncodingsMixin
-        CC_BASES.append(EncodingsMixin)
-    if server_features.audio and server_features.av_sync:
-        from xpra.server.source.avsync_mixin import AVSyncMixin
-        CC_BASES.append(AVSyncMixin)
-from xpra.server.source.idle_mixin import IdleMixin
-CC_BASES.append(IdleMixin)
-CC_BASES = tuple(CC_BASES)
-ClientConnectionClass = type('ClientConnectionClass', CC_BASES, {})
-
 log = Logger("server")
-elog = Logger("encoding")
-keylog = Logger("keyboard")
-mouselog = Logger("mouse")
-timeoutlog = Logger("timeout")
-proxylog = Logger("proxy")
-statslog = Logger("stats")
 notifylog = Logger("notify")
-netlog = Logger("network")
 bandwidthlog = Logger("bandwidth")
 
-log("ClientConnectionClass%s", CC_BASES)
-
 
 BANDWIDTH_DETECTION = envbool("XPRA_BANDWIDTH_DETECTION", True)
 MIN_BANDWIDTH = envint("XPRA_MIN_BANDWIDTH", 5*1024*1024)
@@ -86,28 +34,26 @@ YIELD = envbool("XPRA_YIELD", False)
 counter = AtomicInteger()
 
 
-"""
-This class mediates between the server class (which only knows about actual window objects and display server events)
-and the client specific WindowSource instances (which only know about window ids
-and manage window pixel compression).
-It sends messages to the client via its 'protocol' instance (the network connection),
-directly for a number of cases (cursor, sound, notifications, etc)
-or on behalf of the window sources for pixel data.
+class ClientConnection(StubSourceMixin):
+    """
+    This class mediates between the server class (which only knows about actual window objects and display server events)
+    and the client specific WindowSource instances (which only know about window ids
+    and manage window pixel compression).
+    It sends messages to the client via its 'protocol' instance (the network connection),
+    directly for a number of cases (cursor, sound, notifications, etc)
+    or on behalf of the window sources for pixel data.
 
-Strategy: if we have 'ordinary_packets' to send, send those.
-When we don't, then send packets from the 'packet_queue'. (compressed pixels or clipboard data)
-See 'next_packet'.
+    Strategy: if we have 'ordinary_packets' to send, send those.
+    When we don't, then send packets from the 'packet_queue'. (compressed pixels or clipboard data)
+    See 'next_packet'.
 
-The UI thread calls damage(), which goes into WindowSource and eventually (batching may be involved)
-adds the damage pixels ready for processing to the encode_work_queue,
-items are picked off by the separate 'encode' thread (see 'encode_loop')
-and added to the damage_packet_queue.
-"""
+    The UI thread calls damage(), which goes into WindowSource and eventually (batching may be involved)
+    adds the damage pixels ready for processing to the encode_work_queue,
+    items are picked off by the separate 'encode' thread (see 'encode_loop')
+    and added to the damage_packet_queue.
+    """
 
-class ClientConnection(ClientConnectionClass):
-
-    def __init__(self, protocol, disconnect_cb, session_name, server,
-                 idle_add, timeout_add, source_remove,
+    def __init__(self, protocol, disconnect_cb, session_name,
                  setting_changed,
                  socket_dir, unix_socket_paths, log_disconnect, bandwidth_limit, bandwidth_detection,
                  ):
@@ -119,16 +65,6 @@ class ClientConnection(ClientConnectionClass):
         self.disconnect = disconnect_cb
         self.session_name = session_name
 
-        for bc in CC_BASES:
-            try:
-                bc.__init__(self)
-                bc.init_from(self, protocol, server)
-            except Exception as e:
-                raise Exception("failed to initialize %s: %s" % (bc, e))
-
-        for c in CC_BASES:
-            c.init_state(self)
-
         #holds actual packets ready for sending (already encoded)
         #these packets are picked off by the "protocol" via 'next_packet()'
         #format: packet, wid, pixels, start_send_cb, end_send_cb
@@ -145,34 +81,23 @@ class ClientConnection(ClientConnectionClass):
         self.socket_dir = socket_dir
         self.unix_socket_paths = unix_socket_paths
         self.log_disconnect = log_disconnect
-        self.idle_add = idle_add
-        self.timeout_add = timeout_add
-        self.source_remove = source_remove
 
         self.setting_changed = setting_changed
         # network constraints:
         self.server_bandwidth_limit = bandwidth_limit
         self.bandwidth_detection = bandwidth_detection
 
-        #these statistics are shared by all WindowSource instances:
-        self.statistics = GlobalPerformanceStatistics()
-
-        self.init()
-
+    def run(self):
         # ready for processing:
         self.queue_encode = self.start_queue_encode
-        protocol.set_packet_source(self.next_packet)
+        self.protocol.set_packet_source(self.next_packet)
 
-
-    def __repr__(self):
+    def __repr__(self) -> str:
         return  "%s(%i : %s)" % (type(self).__name__, self.counter, self.protocol)
 
-    def init(self):
+    def init_state(self):
         self.hello_sent = False
         self.info_namespace = False
-        self.send_notifications = False
-        self.send_notifications_actions = False
-        self.notification_callbacks = {}
         self.share = False
         self.lock = False
         self.control_commands = ()
@@ -181,6 +106,9 @@ class ClientConnection(ClientConnectionClass):
         self.soft_bandwidth_limit = self.bandwidth_limit
         self.bandwidth_warnings = True
         self.bandwidth_warning_time = 0
+        self.client_connection_data = {}
+        self.adapter_type = ""
+        self.jitter = 0
         #what we send back in hello packet:
         self.ui_client = True
         self.wants_aliases = True
@@ -190,17 +118,27 @@ class ClientConnection(ClientConnectionClass):
         self.wants_display = True
         self.wants_events = False
         self.wants_default_cursor = False
+        #these statistics are shared by all WindowSource instances:
+        self.statistics = GlobalPerformanceStatistics()
 
 
-    def is_closed(self):
+    def is_closed(self) -> bool:
         return self.close_event.isSet()
 
-    def close(self):
+    def cleanup(self):
         log("%s.close()", self)
-        for c in CC_BASES:
-            c.cleanup(self)
         self.close_event.set()
         self.protocol = None
+        self.statistics.reset(0)
+
+
+    def may_notify(self, *args, **kwargs):
+        #fugly workaround,
+        #MRO is depth first and would hit the default implementation
+        #instead of the mixin unless we force it:
+        notification_mixin = sys.modules.get("xpra.server.source.notification_mixin")
+        if notification_mixin and isinstance(self, notification_mixin.NotificationMixin):
+            notification_mixin.NotificationMixin.may_notify(self, *args, **kwargs)
 
 
     def compressed_wrapper(self, datatype, data, min_saving=128):
@@ -258,27 +196,12 @@ class ClientConnection(ClientConnectionClass):
                 ws.bandwidth_limit = max(MIN_BANDWIDTH//10, bandwidth_limit*weight//total_weight)
 
 
-    def parse_hello(self, c):
-        self.ui_client = c.boolget("ui_client", True)
-        self.wants_encodings = c.boolget("wants_encodings", self.ui_client)
-        self.wants_display = c.boolget("wants_display", self.ui_client)
-        self.wants_events = c.boolget("wants_events", False)
-        self.wants_aliases = c.boolget("wants_aliases", True)
-        self.wants_versions = c.boolget("wants_versions", True)
-        self.wants_features = c.boolget("wants_features", True)
-        self.wants_default_cursor = c.boolget("wants_default_cursor", False)
-
-        for mixin in CC_BASES:
-            mixin.parse_client_caps(self, c)
-
+    def parse_client_caps(self, c : typedict):
         #general features:
         self.info_namespace = c.boolget("info-namespace")
-        self.send_notifications = c.boolget("notifications")
-        self.send_notifications_actions = c.boolget("notifications.actions")
-        log("notifications=%s, actions=%s", self.send_notifications, self.send_notifications_actions)
         self.share = c.boolget("share")
         self.lock = c.boolget("lock")
-        self.control_commands = c.strlistget("control_commands")
+        self.control_commands = c.strtupleget("control_commands")
         self.xdg_menu_update = c.boolget("xdg-menu-update")
         bandwidth_limit = c.intget("bandwidth-limit", 0)
         server_bandwidth_limit = self.server_bandwidth_limit
@@ -288,28 +211,17 @@ class ClientConnection(ClientConnectionClass):
         if self.bandwidth_detection:
             self.bandwidth_detection = c.boolget("bandwidth-detection", True)
         self.client_connection_data = c.dictget("connection-data", {})
-        self.jitter = typedict(self.client_connection_data).intget("jitter", 0)
+        ccd = typedict(self.client_connection_data)
+        self.adapter_type = ccd.strget("adapter-type", "")
+        self.jitter = ccd.intget("jitter", 0)
         bandwidthlog("server bandwidth-limit=%s, client bandwidth-limit=%s, value=%s, detection=%s",
                      server_bandwidth_limit, bandwidth_limit, self.bandwidth_limit, self.bandwidth_detection)
 
-        cinfo = self.get_connect_info()
-        for i,ci in enumerate(cinfo):
-            log.info("%s%s", ["", " "][int(i>0)], ci)
-        if self.client_proxy:
-            from xpra.version_util import version_compat_check
-            msg = version_compat_check(self.proxy_version)
-            if msg:
-                proxylog.warn("Warning: proxy version may not be compatible: %s", msg)
         if getattr(self, "mmap_size", 0)>0:
             log("mmap enabled, ignoring bandwidth-limit")
             self.bandwidth_limit = 0
-        #adjust max packet size if file transfers are enabled:
-        #TODO: belongs in mixin:
-        file_transfer = getattr(self, "file_transfer", None)
-        if file_transfer:
-            self.protocol.max_packet_size = max(self.protocol.max_packet_size, self.file_size_limit*1024*1024)
 
-    def get_socket_bandwidth_limit(self):
+    def get_socket_bandwidth_limit(self) -> int:
         p = self.protocol
         if not p:
             return 0
@@ -341,7 +253,7 @@ class ClientConnection(ClientConnectionClass):
         self.queue_encode(item)
         self.encode_thread = start_thread(self.encode_loop, "encode")
 
-    def encode_queue_size(self):
+    def encode_queue_size(self) -> int:
         ewq = self.encode_work_queue
         if ewq is None:
             return 0
@@ -431,20 +343,13 @@ class ClientConnection(ClientConnectionClass):
 
     def send_async(self, *parts, **kwargs):
         kwargs["synchronous"] = False
+        kwargs["will_have_more"] = False
         self.send(*parts, **kwargs)
 
 
-    def send_hello(self, server_capabilities):
-        capabilities = server_capabilities.copy()
-        for bc in CC_BASES:
-            merge_dicts(capabilities, bc.get_caps(self))
-        self.send("hello", capabilities)
-        self.hello_sent = True
-
-
     ######################################################################
     # info:
-    def get_info(self):
+    def get_info(self) -> dict:
         info = {
                 "protocol"          : "xpra",
                 "connection_time"   : int(self.connection_time),
@@ -452,6 +357,7 @@ class ClientConnection(ClientConnectionClass):
                 "counter"           : self.counter,
                 "hello-sent"        : self.hello_sent,
                 "jitter"            : self.jitter,
+                "adapter-type"      : self.adapter_type,
                 "bandwidth-limit"   : {
                     "detection"     : self.bandwidth_detection,
                     "actual"        : self.soft_bandwidth_limit or 0,
@@ -463,33 +369,22 @@ class ClientConnection(ClientConnectionClass):
                          "connection"       : p.get_info(),
                          })
         info.update(self.get_features_info())
-        for bc in CC_BASES:
-            try:
-                merge_dicts(info, bc.get_info(self))
-            except Exception as e:
-                log("merge_dicts on %s", bc, exc_info=True)
-                log.error("Error: cannot add information from %s:", bc)
-                log.error(" %s", e)
         return info
 
-    def get_features_info(self):
+    def get_features_info(self) -> dict:
         info = {
             "lock"  : bool(self.lock),
             "share" : bool(self.share),
-            "notifications" : self.send_notifications,
             }
         return info
 
 
     def send_info_response(self, info):
-        if self.info_namespace:
-            v = notypedict(info)
-        else:
-            v = flatten_dict(info)
-        self.send_async("info-response", v)
+        self.send_async("info-response", notypedict(info))
 
 
     def send_setting_change(self, setting, value):
+        #we always subclass InfoMixin which defines "client_setting_change":
         if self.client_setting_change:
             self.send_more("setting-change", setting, value)
 
@@ -499,57 +394,7 @@ class ClientConnection(ClientConnectionClass):
             self.send_more("server-event", *args)
 
 
-    ######################################################################
-    # notifications:
-    # Utility functions for mixins (makes notifications optional)
-    def may_notify(self, nid, summary, body, actions=(), hints=None, expire_timeout=10*1000,
-                   icon_name=None, user_callback=None):
-        try:
-            from xpra.platform.paths import get_icon_filename
-            from xpra.notifications.common import parse_image_path
-        except ImportError as e:
-            notifylog("not sending notification: %s", e)
-        else:
-            icon_filename = get_icon_filename(icon_name)
-            icon = parse_image_path(icon_filename) or ""
-            self.notify("", nid, "Xpra", 0, "",
-                        summary, body, actions, hints or {},
-                        expire_timeout, icon, user_callback)
-
-    def notify(self, dbus_id, nid, app_name, replaces_nid, app_icon,
-               summary, body, actions, hints, expire_timeout, icon, user_callback=None):
-        args = (dbus_id, nid, app_name, replaces_nid, app_icon, summary, body, actions, hints, expire_timeout, icon)
-        notifylog("notify%s types=%s", args, tuple(type(x) for x in args))
-        if not self.send_notifications:
-            notifylog("client %s does not support notifications", self)
-            return False
-        if self.suspended:
-            notifylog("client %s is suspended, notification not sent", self)
-            return False
-        if user_callback:
-            self.notification_callbacks[nid] = user_callback
-        #this is one of the few places where we actually do care about character encoding:
-        try:
-            summary = summary.encode("utf8")
-        except UnicodeEncodeError:
-            summary = str(summary)
-        try:
-            body = body.encode("utf8")
-        except UnicodeEncodeError:
-            body = str(body)
-        if self.hello_sent:
-            #Warning: actions and hints are send last because they were added later (in version 2.3)
-            self.send_async("notify_show", dbus_id, nid, app_name, replaces_nid, app_icon,
-                            summary, body, expire_timeout, icon, actions, hints)
-        return True
-
-    def notify_close(self, nid):
-        if not self.send_notifications or self.suspended  or not self.hello_sent:
-            return
-        self.send_more("notify_close", nid)
-
-
-    def set_deflate(self, level):
+    def set_deflate(self, level : int):
         self.send("set_deflate", level)
 
 
